/*
**		Copyright (C) Roman Ukhov 2000
**		e-mail: r_ukhov@yahoo.com
**		http://rukhov.newmail.ru
*/

#include "http_server.hpp"
#include "async_socket.hpp"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

namespace network
{
	async_socket::async_socket(unsigned long _timeout):
	the_socket(INVALID_SOCKET),
	hEvent(WSA_INVALID_EVENT),
	hStopEvent(WSA_INVALID_EVENT),
	l_events(0),
	b_auto_delete(false),
	timeout(_timeout)
	{
		hStopEvent=WSACreateEvent();
		hEvent=WSACreateEvent();

		readO.Offset=0;
		readO.OffsetHigh=0;
		readO.hEvent=CreateEvent(NULL, TRUE, FALSE, NULL);
		
		writeO.Offset=0;
		writeO.OffsetHigh=0;
		writeO.hEvent=CreateEvent(NULL, TRUE, FALSE, NULL);
	}
	
	async_socket::~async_socket()
	{
		WSASetEvent(hStopEvent);

		socket_thread.wait_for_end();

		if(the_socket!=INVALID_SOCKET){
			shutdown();
			::closesocket(the_socket);
			the_socket=INVALID_SOCKET;
		}

		if(hEvent!=WSA_INVALID_EVENT){
			WSACloseEvent(hEvent);
			hEvent=WSA_INVALID_EVENT;
		}
		
		if(hStopEvent!=WSA_INVALID_EVENT){
			WSACloseEvent(hStopEvent);
			hStopEvent=WSA_INVALID_EVENT;
		}
		if(readO.hEvent!=NULL)
			CloseHandle(readO.hEvent);
		if(writeO.hEvent!=NULL)
			CloseHandle(writeO.hEvent);
	}

	bool 
		async_socket::bind(const ip_addr& addr)
	{
		::bind(the_socket, &addr, sizeof(ip_addr));
		return true;
	}

	void 
		async_socket::close()
	{
		WSAResetEvent(hEvent);
		WSASetEvent(hStopEvent);
	}

	bool 
		async_socket::init()
	{
		if(the_socket==INVALID_SOCKET)
			return false;


		WSAResetEvent(hStopEvent);

		return socket_thread.create(static_thread_func, this);
	}

	bool 
		async_socket::create(
		int nSocketType/* = SOCK_STREAM*/,
		long lEvent/* = FD_READ | FD_WRITE | FD_OOB | FD_ACCEPT | FD_CONNECT | FD_CLOSE*/,
		const ip_addr& addr/* =  ip_addr(0,0,0,0,0)*/)
	{
		bool ret_val(true);

		the_socket=socket( AF_INET, nSocketType, 0);

		if(the_socket!=INVALID_SOCKET){
			if(select(lEvent)==true){
				if(bind(addr)==true)
					if(init()==true)
						return true;
			}
		}

		close();

		return false;
	}

	bool 
		async_socket::select(long lEvent/* = FD_READ | FD_WRITE | FD_OOB | FD_ACCEPT | FD_CONNECT | FD_CLOSE */)
	{
		if(WSAEventSelect(  
			the_socket,
			hEvent,
			lEvent)==SOCKET_ERROR){
			return false;
		}else{
			l_events=lEvent;
			return true;
		}
	}

	int 
		async_socket::static_thread_func(void* p)
	{
		return reinterpret_cast<async_socket*>(p)->socket_func();
	}
	
	int 
		async_socket::socket_func()
	{
		WSAEVENT handles[]={
			hStopEvent,
			readO.hEvent,
			writeO.hEvent,
			hEvent};
		WSANETWORKEVENTS netEvents;
		DWORD wait_res;
		DWORD dwTimeOUT;

		on_attach();

		for(;;){

			if(timeout == 0)
				dwTimeOUT=WSA_INFINITE;
			else
				dwTimeOUT=timeout;

			wait_res=WSAWaitForMultipleEvents(
				sizeof(handles)/sizeof(WSAEVENT),
				handles,
				FALSE,
				dwTimeOUT,
				TRUE);
			
			if(wait_res == WSA_WAIT_EVENT_0+3){
				if(WSAEnumNetworkEvents(
					the_socket,
					hEvent,
					&netEvents)!=SOCKET_ERROR){
					if((netEvents.lNetworkEvents&FD_READ) == FD_READ){
						_on_read(netEvents.iErrorCode[FD_READ_BIT]);
					}
					if((netEvents.lNetworkEvents&FD_WRITE) == FD_WRITE){
						_on_write(netEvents.iErrorCode[FD_WRITE_BIT]);
					}
					if((netEvents.lNetworkEvents&FD_CLOSE) == FD_CLOSE){
						_on_close(netEvents.iErrorCode[FD_CLOSE_BIT]);
					}
					if((netEvents.lNetworkEvents&FD_CONNECT) == FD_CONNECT){
						_on_connect(netEvents.iErrorCode[FD_CONNECT_BIT]);
					}
					if((netEvents.lNetworkEvents&FD_ACCEPT) == FD_ACCEPT){
						_on_accept(netEvents.iErrorCode[FD_ACCEPT_BIT]);
					}
				}else{
					int err=WSAGetLastError();
					if(err == WSANOTINITIALISED ||
						err == WSAENOTSOCK){
						close();
					}
				}
			}else if(wait_res == WSA_WAIT_EVENT_0){
				break;
			}else if(wait_res == WSA_WAIT_EVENT_0+1){
				_on_read_complete();
			}else if(wait_res == WSA_WAIT_EVENT_0+2){
				_on_write_complete();
			}else if(wait_res == WSA_WAIT_FAILED){
				break;
			}else if(wait_res == WSA_WAIT_TIMEOUT){
				_on_timeout();
			}else{
				close();
			}
		}

		_on_clear();

		if(b_auto_delete == true)
			delete this;

		return 0;
	}

	void 
		async_socket::_on_close(int err)
	{
		on_close(err);
	}
	
	void 
		async_socket::_on_connect(int err)
	{
	}
	
	void 
		async_socket::_on_read(int err)
	{
		on_read(err);
	}

	void 
		async_socket::_on_write(int err)
	{
		on_write(err);
	}

	void 
		async_socket::_on_accept(int err)
	{
		on_accept(err);
	}

	void 
		async_socket::on_close(int err)
	{

	}

	void 
		async_socket::on_accept(int err)
	{

	}

	void 
		async_socket::on_read(int err)
	{

	}

	void 
		async_socket::on_write(int err)
	{

	}

	int 
		async_socket::get_last_error() const
	{
		return WSAGetLastError();
	}

	bool 
		async_socket::shutdown(int how/* = SD_BOTH*/)
	{
		if(::shutdown(the_socket, how) == SOCKET_ERROR)
			return false;
		else
			return true;
	}

	bool 
		async_socket::listen(int backlog/* = SOMAXCONN*/)
	{
		if(::listen(the_socket, backlog) == SOCKET_ERROR)
			return false;
		else
			return true;
	}

	bool 
		async_socket::attach(SOCKET s, const network::ip_addr& _peername/* = network::ip_addr()*/)
	{
		if(s!=INVALID_SOCKET){
			
			close();

			peername=_peername;
			
			the_socket=s;

			if(init()==true){
				return true;
			}
		}

		return false;
	}

	void 
		async_socket::on_attach()
	{

	}

	async_socket::accepted_socket  
		async_socket::accept() const{
		
		accepted_socket as;
		int len=sizeof(network::ip_addr);

		as.s=::WSAAccept(
				the_socket,
				&as.addr,
				&len,
				NULL, 0);
		return as;
	}

	void 
		async_socket::_on_read_complete()
	{
		DWORD len=0;
		DWORD err=0;

		if(GetOverlappedResult(
			(HANDLE)the_socket,
			&readO,
			&len,
			FALSE)==FALSE){
			err=GetLastError();
			len=0;
		}

		ResetEvent(readO.hEvent);
		on_read_complete(err, len);
	}

	void 
		async_socket::_on_write_complete()
	{
		DWORD err=0;
		DWORD len;

		if(GetOverlappedResult(
			(HANDLE)the_socket,
			&writeO,
			&len,
			FALSE)==FALSE){
			err=GetLastError();
			len=0;
		}
		
		ResetEvent(writeO.hEvent);
		on_write_complete(err, len);
	}

	bool 
		async_socket::read(char* pBuff, DWORD len)
	{
		DWORD readed;

		if(ReadFile(
			(HANDLE) the_socket,
			pBuff,
			len,
			&readed,
			&readO) == FALSE){
			DWORD err = GetLastError();

			if(err == ERROR_IO_PENDING)
				return true;
			else
				return false;
		}else
			return true;
	}

	bool 
		async_socket::write(char* pBuff, DWORD len)
	{
		DWORD writeded;

		if(WriteFile(
			(HANDLE) the_socket,
			pBuff,
			len,
			&writeded,
			&writeO) == FALSE){
			DWORD err = GetLastError();

			if(err == ERROR_IO_PENDING)
				return true;
			else
				return false;
		}else
			return true;
	}

	void 
		async_socket::on_read_complete(DWORD err, DWORD len)
	{

	}

	void 
		async_socket::on_write_complete(DWORD err, DWORD len)
	{

	}

	void 
		async_socket::on_clear()
	{
	}

	void 
		async_socket::set_priority(int pr/* = THREAD_PRIORITY_NORMAL */)
	{
		socket_thread.set_priority(pr);
	}

	void 
		async_socket::_on_clear()
	{
		if(the_socket!=INVALID_SOCKET){
			select(0);
			WSANETWORKEVENTS netEvents;
			if(WSAEnumNetworkEvents(
				the_socket,
				hEvent,
				&netEvents)!=SOCKET_ERROR){
				select(0);
			}

			select(0);
			shutdown();
			::closesocket(the_socket);
			the_socket=INVALID_SOCKET;
		}

		on_clear();
	}

	void 
		async_socket::_on_timeout()
	{
		on_timeout();
	}

	void 
		async_socket::on_timeout()
	{
	}
}

